#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <gtk--.h>
#include <gdk--/color.h>
#include <linux/isdn.h>
#include <limits.h>

#define APP_NAME "Gtk ISDN Monitor"
#define IM_BUFSIZE (10+(22*ISDN_MAX_CHANNELS))
#define MAX_TODELETE (ISDN_MAX_CHANNELS*5)

char *server = 0;
char *devicePath = 0;
bool remoteEnable = false;

class lpMessageDialog : public Gtk_Window
 {
  public:
   lpMessageDialog(char *msg, char *title = APP_NAME);
   
  protected:
   Gtk_Label messageLabel;
   Gtk_Viewport viewport; 
   Gtk_VBox vbox;
   Gtk_Button okButton;
   
   void onOkButtonClicked();
 };

lpMessageDialog::lpMessageDialog(char *msg, char *title)
 : messageLabel(msg), okButton("OK")
 {
  set_title(title);
  set_border_width(5);

  messageLabel.set_padding(15, 15);
  viewport.add(messageLabel);
  vbox.pack_start(viewport, true, true, 2);

  connect_to_method(okButton.clicked, this, &lpMessageDialog::onOkButtonClicked);
  vbox.pack_end(okButton, false, false, 2);
  
  vbox.show_all();
  add(vbox);
  
  set_modal(true);
 }

void lpMessageDialog::onOkButtonClicked()
 {
  delete_self();
 }
 
class lpConfigDialog : public Gtk_Window
 {
  public :
   lpConfigDialog();
  
  protected:
   Gtk_VBox entireVBox, configVBox;
   Gtk_HBox buttonHBox, serverHBox, fileHBox;
   Gtk_Button okButton, cancelButton;
   Gtk_CheckButton serverCheckButton;
   Gtk_Entry serverEntry, fileEntry;
   Gtk_Label serverLabel, fileLabel;
   Gtk_Frame serverFrame, fileFrame;

   void onOkButtonClicked();
   void onCancelButtonClicked();
   void onServerCheckButtonClicked();
 };
 

class lpMainWindow: public Gtk_Window
 {
  public:
   lpMainWindow();
   virtual ~lpMainWindow();

  protected:
   Gtk_VBox vbox;
   Gtk_HBox hbox;
   Gtk_Label label;
   Gtk_CList clist;
   Gtk_Button closeButton, configButton;
   Gtk_Viewport viewport1, viewport2, viewport3;
   
   gint delete_event_impl(GdkEventAny*);
   void onCloseButtonClicked();
   void onConfigButtonClicked();
   gint onTimeout();
  
   
   FILE *f;
   
   char* toDelete[MAX_TODELETE];
   int tdCount;
   
   lpConfigDialog *configDialog;
   
   void saveOptions();
   void loadOptions();
   
  public: 
   void openISDNInfo();
   void updateContents();
 };

lpMainWindow *mainWindow;

lpConfigDialog::lpConfigDialog()
 : Gtk_Window(GTK_WINDOW_DIALOG), serverCheckButton("Enable remote ISDN monitoring"), serverLabel("Server:"), okButton("OK"), cancelButton("Cancel"), serverFrame("Remote"), fileLabel("File/Device:"), fileFrame("Local")
 {
  set_title(APP_NAME);
  set_border_width(5);

  entireVBox.set_spacing(10);

  connect_to_method(serverCheckButton.clicked, this, &lpConfigDialog::onServerCheckButtonClicked);
  entireVBox.pack_start(serverCheckButton, false, false, 2);
  
  fileHBox.pack_start(fileLabel, false, false, 2);
  fileHBox.pack_end(fileEntry, true, true, 2);
  fileHBox.set_border_width(10);
  fileHBox.set_spacing(5);
  fileFrame.add(fileHBox);
  entireVBox.pack_start(fileFrame, false, false, 2);
  
  serverHBox.pack_start(serverLabel, false, false, 2);
  serverHBox.pack_end(serverEntry, true, true, 2);
  serverHBox.set_border_width(10);
  serverHBox.set_spacing(5);
  serverFrame.add(serverHBox);
  entireVBox.pack_start(serverFrame, false, false, 2);

  connect_to_method(okButton.clicked, this, &lpConfigDialog::onOkButtonClicked);
  connect_to_method(cancelButton.clicked, this, &lpConfigDialog::onCancelButtonClicked);

  buttonHBox.pack_start(okButton, true, true, 2);
  buttonHBox.pack_start(cancelButton, true, true, 2);
  buttonHBox.set_homogeneous(true);
  entireVBox.pack_end(buttonHBox, false, false, 2);
  
  
  entireVBox.show_all();
  add(entireVBox);
  set_modal(true);
  
  serverCheckButton.set_active(remoteEnable);
  serverEntry.set_text(server ? server : "localhost");
  fileEntry.set_text(devicePath ? devicePath : "/dev/isdninfo");
  
  onServerCheckButtonClicked();
 }

void lpConfigDialog::onOkButtonClicked()
 {
  remoteEnable = serverCheckButton.get_active();

  if (server) 
   {
    free(server);
    server = 0;
   }    
    
  server = strdup(serverEntry.get_text().c_str());
  devicePath = strdup(fileEntry.get_text().c_str());
  
  mainWindow->openISDNInfo();
  mainWindow->updateContents();
  
  hide();
 }
 
void lpConfigDialog::onCancelButtonClicked()
 {
  hide();
 }

void lpConfigDialog::onServerCheckButtonClicked()
 {
  bool b = serverCheckButton.get_active();
 
  serverLabel.set_sensitive(b);
  serverEntry.set_sensitive(b);
  serverFrame.set_sensitive(b);

  fileLabel.set_sensitive(!b);
  fileEntry.set_sensitive(!b);
  fileFrame.set_sensitive(!b);
 }


 
lpMainWindow::lpMainWindow()
 : Gtk_Window(GTK_WINDOW_TOPLEVEL), label(APP_NAME), clist(5), closeButton("Close"), configButton("Configure..."), tdCount(0)
 {
  mainWindow = this;
  configDialog = 0;
  
  set_title(APP_NAME);
  set_border_width(5);

  vbox.set_spacing(10);

  label.set_padding(5, 5);
  
  viewport1.add(label);
  vbox.pack_start(viewport1, false, false, 0);
  
  clist.set_usize(300, 150);
  clist.set_column_title(0, "#");
  clist.set_column_title(1, "Line");
  clist.set_column_title(2, "Usage");
  clist.set_column_title(3, "Type");
  clist.set_column_width(3, 50);
  clist.set_column_title(4, "Number");
  clist.column_titles_show();
  clist.set_shadow_type(GTK_SHADOW_IN);
//  clist.set_selection_mode(GTK_SELECTION_BROWSE);
  clist.column_titles_passive();
  clist.set_column_auto_resize(0, 1);
  clist.set_column_auto_resize(1, 1);
  clist.set_column_auto_resize(2, 1);
  clist.set_column_auto_resize(3, 1);
  clist.set_column_auto_resize(4, 1);
  vbox.pack_start(clist, true, true, 0);

  connect_to_method(configButton.clicked, this, &lpMainWindow::onConfigButtonClicked);

  //  connect_to_method(okButton.clicked, this, &lpConfigDialog::onOkButtonClicked);

  viewport2.add(configButton);
  connect_to_method(closeButton.clicked, this, &lpMainWindow::onCloseButtonClicked);
  viewport3.add(closeButton);
  hbox.pack_start(viewport2, true, true, 0);
  hbox.pack_end(viewport3, true, true, 0);
  hbox.set_homogeneous(true);
  hbox.set_spacing(10);
  
  vbox.pack_end(hbox, false, false, 0);
    
  add(vbox);
  vbox.show_all();

  f = 0;
  server = 0;
  devicePath = 0;
  remoteEnable = 0;
  
  loadOptions();
  openISDNInfo();
  updateContents();
  
  connect_to_method(Gtk_Main::timeout(500), this, &lpMainWindow::onTimeout);
 }

lpMainWindow::~lpMainWindow()
 {
  if (f)
   fclose(f);
   
  saveOptions(); 
  
  for (int i = 0; i < tdCount; i++)
   if (toDelete[i])
    free(toDelete[i]);
    
  tdCount = 0;  
  
  if (configDialog)
   {
    delete configDialog;
   }    
   
  if (server)
   free(server);
 }

gint lpMainWindow::delete_event_impl(GdkEventAny*)
 {
  Gtk_Main::instance()->quit();
  return 0;
 }

void lpMainWindow::onCloseButtonClicked()
 {
  Gtk_Main::instance()->quit();
 }

void lpMainWindow::onConfigButtonClicked()
 {
  if (configDialog)
   {
    if (!configDialog->is_visible())
     {
      delete configDialog;
      configDialog = 0;
     }
    else
     configDialog->activate(); 
   }
   
  if (!configDialog)
   {
    configDialog = new lpConfigDialog();
    configDialog->show();
   }
 } 
 
void parseString(char *buf, char* v[], int l)
 {
  char *p;
  
  p = strtok(buf, "\t");
  for (int i = 0; i < l; i++)
   {
    p = strtok(0, " ");
    if (!p) break;
    v[i] = strdup(p);
   }
 }

void parseInt(char *buf, int v[], int l)
 {
  char *p;
  
  p = strtok(buf, "\t");
  for (int i = 0; i < l; i++)
   {
    p = strtok(0, " ");
    if (!p) break;
    v[i] = atoi(p);
   }
 }

 
void lpMainWindow::updateContents()
 {
  if (f)
   {
    fd_set fs;
    timeval tv;

    FD_ZERO(&fs);
    FD_SET(fileno(f), &fs);
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    
    if ((select(FD_SETSIZE, &fs, 0, 0, &tv) > 0) && (!feof(f)))
     {
      char *idmap[ISDN_MAX_CHANNELS];
      int chmap[ISDN_MAX_CHANNELS];  
      int drmap[ISDN_MAX_CHANNELS];
      int usage[ISDN_MAX_CHANNELS];
      int flags[ISDN_MAX_CHANNELS];
      char *phone[ISDN_MAX_CHANNELS];
      char buf[IM_BUFSIZE];

      fgets(buf, sizeof(buf), f);
      parseString(buf, idmap, ISDN_MAX_CHANNELS);
      fgets(buf, sizeof(buf), f);
      parseInt(buf, chmap, ISDN_MAX_CHANNELS);
      fgets(buf, sizeof(buf), f);
      parseInt(buf, drmap, ISDN_MAX_CHANNELS);
      fgets(buf, sizeof(buf), f);
      parseInt(buf, usage, ISDN_MAX_CHANNELS);
      fgets(buf, sizeof(buf), f);
      parseInt(buf, flags, ISDN_MAX_CHANNELS);
      fgets(buf, sizeof(buf), f);
      parseString(buf, phone, ISDN_MAX_CHANNELS);

      clist.clear();
      
      for (int i = 0; i < tdCount; i++)
       if (toDelete[i])
        free(toDelete[i]);

      tdCount = 0;  
      
      clist.thaw();
      
      for (int i = 0; i < ISDN_MAX_CHANNELS; i++)
       {
        gchar *entry[5];
        char s[256];
      
        if (chmap[i] < 0) break;
      
        sprintf(s, "%i", chmap[i]);
        entry[0] = strdup(s);
        toDelete[tdCount++] = entry[0];
      
        entry[1] = strdup(idmap[i]);
        toDelete[tdCount++] = entry[1];
      
        if (usage[i] & ISDN_USAGE_OUTGOING)
         strcpy(s, "Out");
        else if (usage[i] == ISDN_USAGE_EXCLUSIVE)
         strcpy(s, "Excl.");
        else if (usage[i])
         strcpy(s, "In");
        else
         strcpy(s, "Off");
	 
        entry[2] = strdup(s);       
        toDelete[tdCount++] = entry[2];
	
      
        switch (usage[i] & ISDN_USAGE_MASK)
         {
          case ISDN_USAGE_RAW :
  	   strcpy(s, "Raw");
  	   break;
	  case ISDN_USAGE_MODEM :
	   strcpy(s, "Modem");
	   break;
	  case ISDN_USAGE_NET :
	   strcpy(s, "Net");
	   break;
	  case ISDN_USAGE_VOICE :
	   strcpy(s, "Voice");
	   break;
	  case ISDN_USAGE_FAX :
	   strcpy(s, "Fax");
	   break;
	  case 0 :
	   s[0] = 0;
	   break;
	  default :
	   strcpy(s, "??");
	   break;
         }
        entry[3] = strdup(s);         
        toDelete[tdCount++] = entry[3];
	

        if (usage[i] & ISDN_USAGE_MASK)
         strcpy(s, phone[i]);
        else
         s[0] = 0;
	 
        entry[4] = strdup(s);       
        toDelete[tdCount++] = entry[4];
       
        clist.append((const char**)entry);
       }
       
      for (int i = 0; i < ISDN_MAX_CHANNELS; i++)
       {
        if (phone[i]) free(phone[i]);
        if (idmap[i]) free(idmap[i]);
       }
    
      clist.freeze();
     }
   }    
  else
   {
    clist.clear();
   } 
 }

gint lpMainWindow::onTimeout()
 {
  updateContents();
  return 1;
 }

void lpMainWindow::openISDNInfo()
 {
  if (f)
   {
    fclose(f);
    f = 0;
   }
  
  if (remoteEnable)
   {
    int ourPort = 0;
    char ourHost[256];
    char *p;
    sockaddr_in sai;
    hostent *he;
    
    strcpy(ourHost, server);
    p = strchr(ourHost, ':');
    if (p)
     {
      *p = 0;
      p++;
      ourPort = atoi(p);
     }
        
    if (ourPort == 0)
     ourPort = 4711;
     
//    printf("h: \"%s\"; p: \"%i\"\n", ourHost, ourPort);
     
    int s = socket(PF_INET, SOCK_STREAM, 0);
    if (s)
     {
      he = gethostbyname(ourHost);
      if (he)
       {
        sai.sin_family = AF_INET;  
        sai.sin_port = htons(ourPort);
        sai.sin_addr = *(struct in_addr*) he->h_addr;
	
	if (connect(s, (sockaddr*) &sai, sizeof(sai)) == 0)
	 {
	  f = fdopen(s, "r");
	 }
	else
	 close(s); 
       }	
      else
       close(s); 
     }
   }
  else
   {
    f = fopen(devicePath ? devicePath : "/dev/isdninfo", "r");
   }
   
  if (!f)
   {
    lpMessageDialog *md = new lpMessageDialog("Could not open ISDNINFO!");
    md->show();
   }
 }

void lpMainWindow::saveOptions()
 {
  FILE *f2;
  char fn[PATH_MAX];
  char *p;
  
  p = getenv("HOME");
  if (p) 
   {
    sprintf(fn, "%s/.gisdnrc", p);
    f2 = fopen(fn, "w");
    if (f2)
     {
      fprintf(f2, "%i\n", remoteEnable ? 1 : 0);
      fprintf(f2, "%s\n", server ? server : "");
      fprintf(f2, "%s\n", devicePath ? devicePath : "");
      fclose(f2);
     }
   }
 }
 
void lpMainWindow::loadOptions()
 {
  FILE *f2;
  char fn[PATH_MAX];
  char *p;
  
  p = getenv("HOME");
  if (p) 
   {
    sprintf(fn, "%s/.gisdnrc", p);
    f2 = fopen(fn, "r");
    if (f2)
     {
      char ln[256];
      if (fgets(ln, sizeof(ln), f2) > 0)
       remoteEnable = atoi(ln) ? true : false;
      
      if (fgets(ln, sizeof(ln), f2) > 0)
       {
        if (strlen(ln))
        ln[strlen(ln)-1] = 0;

        if (server) free(server);
        server = strdup(ln);
       }	
      
      if (fgets(ln, sizeof(ln), f2) > 0)
       {
        if (strlen(ln))
         ln[strlen(ln)-1] = 0;

        if (devicePath) free(devicePath);
        devicePath = strdup(ln);
       }	

      fclose(f2);
     }
   }
 }


int main(int argc, char *argv[])
 {
  Gtk_Main myapp(&argc, &argv);
  lpMainWindow mywin;
  
  mywin.show();
  myapp.run();
  
  return 0; 
}
